import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.excaliburgapfillvdsgenerator import ExcaliburGapFillVDSGenerator

vdsgen_patch_path = "vdsgen.excaliburgapfillvdsgenerator"
GapFillVDSGenerator_patch_path = vdsgen_patch_path + ".GapFillVDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class ExcaliburGapFillVDSGeneratorTester(ExcaliburGapFillVDSGenerator):
    """A version of ExcaliburGapFillVDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class ExcaliburGapFillVDSGeneratorInitTest(unittest.TestCase):

    @patch(GapFillVDSGenerator_patch_path + '.__init__')
    def test_super_called(self, super_mock):
        ExcaliburGapFillVDSGenerator("/test/path", files=["raw.h5"])

        super_mock.assert_called_once_with(
            "/test/path", None, ["raw.h5"], None, None, None, None, None,
            256, 256, 8, 2, None)


class SimpleFunctionsTest(unittest.TestCase):

    @patch(GapFillVDSGenerator_patch_path + '.grab_metadata',
           return_value=dict(frames=(3,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = ExcaliburGapFillVDSGeneratorTester(
            files=["raw.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_called_once_with("raw.h5")
        self.assertEqual(expected_source, source)

    def test_construct_vds_spacing(self):
        gen = ExcaliburGapFillVDSGeneratorTester(
            files=[""] * 6,
            grid_x=8, grid_y=6, chip_spacing=10, module_spacing=100)
        expected_x_spacing = [10, 10, 10, 10, 10, 10, 10, 0]
        expected_y_spacing = [10, 100, 10, 100, 10, 0]

        x_spacing, y_spacing = gen.construct_vds_spacing()

        self.assertEqual(expected_x_spacing, x_spacing)
        self.assertEqual(expected_y_spacing, y_spacing)

    file_mock = MagicMock()

    @patch(vdsgen_patch_path +
           '.ExcaliburGapFillVDSGenerator.construct_vds_spacing',
           return_value=([3, 3, 3, 3, 3, 3, 3, 0], [3, 123, 3, 123, 3, 0]))
    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(h5py_patch_path + '.VirtualSource')
    @patch(h5py_patch_path + '.VirtualLayout')
    def test_create_virtual_layout(self, layout_mock, source_mock, file_mock,
                                   construct_mock):
        gen = ExcaliburGapFillVDSGeneratorTester(
            output_file="/test/path/vds.h5",
            grid_x=8, grid_y=6, sub_height=256, sub_width=256,
            source_node="data", target_node="full_frame",
            source_file="raw.h5", name="gaps.h5")
        source = vdsgenerator.SourceMeta(
            frames=(3,), height=1536, width=2048, dtype="uint16")
        dataset_mock = MagicMock()
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.__getitem__.return_value = dataset_mock

        layout = gen.create_virtual_layout(source)

        layout_mock.assert_called_once_with((3, 1791, 2069), "uint16")
        source_mock.assert_called_once_with(
            "raw.h5", name="data", dtype="uint16", shape=(3, 1536, 2048))
        # TODO: Pass numpy arrays to check slicing



"""Make things easy to import."""
from .subframevdsgenerator import SubFrameVDSGenerator
from .interleavevdsgenerator import InterleaveVDSGenerator
from .excaliburgapfillvdsgenerator import ExcaliburGapFillVDSGenerator
from .reshapevdsgenerator import ReshapeVDSGenerator

from .rawsourcegenerator import generate_raw_files

__all__ = ["InterleaveVDSGenerator", "SubFrameVDSGenerator",
           "ReshapeVDSGenerator", "ExcaliburGapFillVDSGenerator",
           "generate_raw_files"]



import unittest
from mock import MagicMock, patch, call

from vdsgen import app

app_patch_path = "vdsgen.app"
parser_patch_path = app_patch_path + ".ArgumentParser"
VDSGenerator_patch_path = app_patch_path + ".VDSGenerator"
InterleaveVDSGenerator_patch_path = app_patch_path + ".InterleaveVDSGenerator"
SubFrameVDSGenerator_patch_path = app_patch_path + ".SubFrameVDSGenerator"
ExcaliburGapFillVDSGenerator_patch_path = app_patch_path + \
                                          ".ExcaliburGapFillVDSGenerator"
ReshapeVDSGenerator_patch_path = app_patch_path + \
                                 ".ReshapeVDSGenerator"


class ParseArgsTest(unittest.TestCase):

    @patch(parser_patch_path + '.error')
    @patch(parser_patch_path + '.parse_args',
           return_value=MagicMock(empty=True, files=None))
    def test_empty_and_not_files_then_error(self, parse_mock, error_mock):

        app.parse_args()

        parse_mock.assert_called_once_with()
        error_mock.assert_called_once_with(
            "To make an empty VDS you must explicitly define --files for the "
            "eventual raw datasets.")

    @patch(parser_patch_path + '.error')
    @patch(parser_patch_path + '.parse_args',
           return_value=MagicMock(mode="gap-fill", files=["one.h5", "two.h5"]))
    def test_gap_fill_only_one_file(self, parse_mock, error_mock):

        app.parse_args()

        parse_mock.assert_called_once_with()
        error_mock.assert_called_once_with(
            "Gap fill can only operate on a single dataset.")


class MainTest(unittest.TestCase):

    @patch(SubFrameVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(mode="sub-frames", empty=True))
    def test_main_empty(self, parse_mock, init_mock):
        gen_mock = init_mock.return_value
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=dict(shape=args_mock.shape, dtype=args_mock.data_type),
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            stripe_spacing=args_mock.stripe_spacing,
            module_spacing=args_mock.module_spacing,
            fill_value=args_mock.fill_value,
            log_level=args_mock.log_level)

        gen_mock.generate_vds.assert_called_once_with()

    @patch(SubFrameVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(mode="sub-frames", empty=False))
    def test_main_not_empty(self, parse_mock, init_mock):
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=None,
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            stripe_spacing=args_mock.stripe_spacing,
            module_spacing=args_mock.module_spacing,
            fill_value=args_mock.fill_value,
            log_level=args_mock.log_level)

    @patch(InterleaveVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(mode="interleave", empty=False))
    def test_main_interleave(self, parse_mock, init_mock):
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=None,
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            block_size=args_mock.block_size,
            fill_value=args_mock.fill_value,
            log_level=args_mock.log_level)

    @patch(ExcaliburGapFillVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(mode="gap-fill", modules=3, empty=False))
    def test_main_gap_fill(self, parse_mock, init_mock):
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=None,
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            modules=args_mock.modules,
            chip_spacing=args_mock.stripe_spacing,
            module_spacing=args_mock.module_spacing,
            fill_value=args_mock.fill_value,
            log_level=args_mock.log_level)

    @patch(ReshapeVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(mode="reshape", empty=False))
    def test_main_reshape(self, parse_mock, init_mock):
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            tuple(args_mock.new_shape),
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=None,
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            fill_value=args_mock.fill_value,
            log_level=args_mock.log_level,
            alternate=args_mock.alternate
        )



import numpy as np
import h5py

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "prefix", type=str, help="Path and name prefix")
    parser.add_argument(
        "frames", type=int, help="Number of frames.")
    parser.add_argument(
        "files", type=int, help="Number of files to spread frames across.")
    parser.add_argument(
        "block_size", type=int, default=1,
        help="Size of contiguous blocks of frames.")
    parser.add_argument(
        "x_dim", type=int, default=100,
        help="Width of frame")
    parser.add_argument(
        "y_dim", type=int, default=100,
        help="Height of frame")
    parser.add_argument(
        "dset", type=str, default="data",
        help="Dataset name")

    return parser.parse_args()


def generate_raw_files(prefix, frames, files, block_size, x_dim, y_dim,
                       dset="data"):

    values = []
    for _ in range(files):
        values.append(list())

    for frame_idx in range(frames):
        if files > 1:
            block_idx = frame_idx // block_size
            file_idx = block_idx % files
        else:
            file_idx = 0
        values[file_idx].append(frame_idx)

    for file_idx, file_values in enumerate(values):
        with h5py.File(prefix + "_{}.h5".format(file_idx)) as f:
            f.create_dataset(dset,
                             shape=(len(values[file_idx]),
                                    y_dim, x_dim))
            for idx, value in enumerate(file_values):
                f[dset][idx] = np.full((y_dim, x_dim),
                                        value, dtype="int8")


def main():
    """Run program."""
    args = parse_args()

    generate_raw_files(args.prefix, args.files, args.frames, args.block_size,
                       args.x_dim, args.y_dim, args.dset)


if __name__ == "__main__":
    sys.exit(main())



"""A class to generate an ND Virtual Dataset from a 1D raw dataset."""

import logging

import h5py as h5

from .vdsgenerator import VDSGenerator, SourceMeta


class ReshapeVDSGenerator(VDSGenerator):
    """A class to generate an ND Virtual Dataset from a 1D raw dataset."""

    logger = logging.getLogger("ReshapeVDSGenerator")

    def __init__(self, shape,
                 path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 log_level=None,
                 alternate=None):
        """
        Args:
            shape(tuple(int)): Shape of output dataset
            alternate(tuple(bool)): Whether each axis alternates

        """
        super(ReshapeVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

        self.total_frames = 0
        self.periods = []
        self.alternate = alternate
        self.dimensions = shape
        self.source_file = self.files[0]  # Reshape only has one raw file

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.files[0])
        self.total_frames = data["frames"][0]
        for dataset in self.files[1:]:
            temp_data = self.grab_metadata(dataset)
            self.total_frames += temp_data["frames"][0]
            for attribute, value in data.items():
                if attribute != "frames" and temp_data[attribute] != value:
                    raise ValueError("Files have mismatched "
                                     "{}".format(attribute))

        source = SourceMeta(frames=data['frames'],
                            height=data['height'], width=data['width'],
                            dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", self.total_frames, *source[1:])
        return source

    def create_virtual_layout(self, source_meta):
        """Create a VirtualLayout mapping raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            VirtualLayout: Object describing links between raw data and VDS

        """
        if source_meta.frames[0] != self.product(self.dimensions):
            raise ValueError(
                "Length of source frames ({}) does no match target shape "
                "[{}] ({}))".format(
                    source_meta.frames[0],
                    ", ".join(str(d) for d in self.dimensions),
                    self.product(self.dimensions))
            )
        vds_shape = self.dimensions + (source_meta.height, source_meta.width)
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n", vds_shape)
        v_layout = h5.VirtualLayout(vds_shape, source_meta.dtype)

        source_shape = source_meta.frames + \
            (source_meta.height, source_meta.width)
        v_source = h5.VirtualSource(
            self.source_file, name=self.source_node,
            shape=source_shape, dtype=source_meta.dtype
        )

        if self.alternate is not None:
            v_layout = self.create_alternating_virtual_layout(
                vds_shape, v_source, v_layout)
        else:
            v_layout[...] = v_source

        return v_layout

    def create_alternating_virtual_layout(self, shape, v_source, v_layout):
        radices = self.create_mixed_radix_set()

        # Iterate over total number of row hyperslabs to map to VDS
        for idx in range(self.product(self.dimensions)):
            axis_indices = self.calculate_axis_indices(idx, radices, shape)
            # Hyperslab: Single index for each inner axis,
            #            Full extent of outermost axis,
            #            Full slice for height and width
            vds_hyperslab = tuple(axis_indices +
                                  [self.FULL_SLICE, self.FULL_SLICE])
            v_layout[vds_hyperslab] = v_source[idx]

            self.logger.debug(
                "Mapping %s[%s, ...] to %s[%d, ...].",
                self.name, ", ".join(str(idx) for idx in axis_indices),
                self.source_file.split("/")[-1], idx)

        return v_layout

    def create_mixed_radix_set(self):
        # Create a mixed radix set mapping any 1D index to an ND index
        # The 1D index is a decimal number and the ND index is the equivalent
        # representation in the mixed radix numeral system derived from shape
        # e.g. for shape (5, 3, 10) radices = 30, 10 and 1
        #   132 -> 412 (30*4 + 10*1 + 1*2) so [132] in 1D -> [4, 1, 2] in 3D
        radices = [1]  # Smallest radix is always worth 1 in decimal
        for axis_length in reversed(self.dimensions[1:]):
            radices.insert(0, radices[0] * axis_length)
        return tuple(radices)

    @staticmethod
    def product(iterable):
        """Calculate product of elements of an iterable.

        Args:
            iterable: An object capable of returning its members one at a time.
                Must have at least on element.

        Returns:
            int: Product

        """
        product = 1
        for value in iterable:
            product *= value
        return product

    def calculate_axis_indices(self, frame_index, radices, shape):
        """Calculate indices for each inner axis for this frame index.

        Args:
            frame_index(int): Frame index in overall dataset
            radices(tuple): Mixed radix numeral definition
            shape(tuple): Shape of dataset

        Returns:
            list: Indices for each individual axis

        """
        axis_indices = [0 for _ in self.dimensions]
        for idx, radix in enumerate(radices):
            while frame_index >= radix:
                frame_index -= radix
                axis_indices[idx] += 1

        # Invert axis indices for axes that are alternating
        for axis in range(1, len(self.dimensions)):
            if self.alternate[axis] and axis_indices[axis - 1] % 2 != 0:
                # If this axis alternates direction and the parent axis index
                # is odd, then invert this axis index
                axis_indices[axis] = shape[axis] - 1 - axis_indices[axis]

        return axis_indices



import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.interleavevdsgenerator import InterleaveVDSGenerator

vdsgen_patch_path = "vdsgen.interleavevdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class InterleaveVDSGeneratorTester(InterleaveVDSGenerator):
    """A version of VDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class FrameVDSGeneratorInitTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.__init__')
    def test_super_called(self, super_mock):
        gen = InterleaveVDSGenerator("/test/path", prefix="stripe_")

        super_mock.assert_called_once_with("/test/path", "stripe_",
                                           *[None]*7)


class SimpleFunctionsTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           return_value=dict(frames=(5,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = InterleaveVDSGeneratorTester(files=["stripe_1.h5", "stripe_2.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(5, 5), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])
        self.assertEqual(expected_source, source)

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           side_effect=[
               dict(frames=(3,), height=256, width=2048, dtype="uint16"),
               dict(frames=(3,), height=512, width=2048, dtype="uint16")])
    def test_process_source_datasets_given_mismatched_data(self, grab_mock):
        gen = InterleaveVDSGeneratorTester(files=["stripe_1.h5", "stripe_2.h5"])

        with self.assertRaises(ValueError):
            gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])

    file_mock = MagicMock()

    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(h5py_patch_path + '.VirtualSource',
           side_effect=[MagicMock(shape=(3, 256, 2048)),
                        MagicMock(shape=(2, 256, 2048))])
    @patch(h5py_patch_path + '.VirtualLayout')
    def test_create_virtual_layout(self, layout_mock, source_mock, file_mock):
        gen = InterleaveVDSGeneratorTester(
            output_file="/test/path/vds.hdf5",
            target_node="full_frame", source_node="data",
            files=["raw1.h5", "raw2.h5"], name="vds.hdf5",
            block_size=1)
        source = vdsgenerator.SourceMeta(
            frames=(3, 2), height=256, width=2048, dtype="uint16")
        dataset_mock = MagicMock()
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.__getitem__.return_value = dataset_mock

        layout = gen.create_virtual_layout(source)

        layout_mock.assert_called_once_with((5, 256, 2048), "uint16")
        source_mock.assert_has_calls(
            [call("raw1.h5", dtype="uint16", name="data", shape=(3, 256, 2048)),
             call("raw2.h5", dtype="uint16", name="data", shape=(2, 256, 2048))],
            any_order=True)
        # TODO: Pass numpy arrays to check slicing



import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.gapfillvdsgenerator import GapFillVDSGenerator

vdsgen_patch_path = "vdsgen.gapfillvdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class GapFillVDSGeneratorTester(GapFillVDSGenerator):
    """A version of GapFillVDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class GapFillVDSGeneratorInitTest(unittest.TestCase):

    super_mock = MagicMock()

    def set_files(self, _path, _prefix, files, *_):
        """A method to patch VDSGenerator __init__ with"""
        # Set self.files so the GapFill __init__ works
        self.files = files
        # Make a call we can test for
        GapFillVDSGeneratorInitTest.super_mock(_path, _prefix, files, *_)

    @classmethod
    def setUp(cls):
        cls.super_mock.reset_mock()

    @patch(VDSGenerator_patch_path + '.__init__', new=set_files)
    def test_init(self):
        GapFillVDSGenerator("/test/path", files=["raw.h5"])

        self.super_mock.assert_called_once_with(
            "/test/path", None, ["raw.h5"], *[None]*6)

    @patch(VDSGenerator_patch_path + '.__init__', new=set_files)
    def test_init_given_more_than_one_file_then_error(self):
        self.assertRaises(ValueError,
                          GapFillVDSGenerator,
                          "/test/path", files=["raw_1.h5", "raw_2.h5"])


class SimpleFunctionsTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           return_value=dict(frames=(3,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = GapFillVDSGeneratorTester(
            files=["raw.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_called_once_with("raw.h5")
        self.assertEqual(expected_source, source)

    def test_construct_vds_spacing(self):
        gen = GapFillVDSGeneratorTester()

        with self.assertRaises(NotImplementedError):
            gen.construct_vds_spacing()

    def test_create_virtual_layout(self):
        gen = GapFillVDSGeneratorTester()

        with self.assertRaises(NotImplementedError):
            gen.create_virtual_layout(MagicMock())



import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.subframevdsgenerator import SubFrameVDSGenerator

vdsgen_patch_path = "vdsgen.subframevdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class SubFrameVDSGeneratorTester(SubFrameVDSGenerator):
    """A version of SubFrameVDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class SubFrameVDSGeneratorInitTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.__init__')
    def test_super_called(self, super_mock):
        SubFrameVDSGenerator("/test/path", prefix="stripe_")

        super_mock.assert_called_once_with("/test/path", "stripe_",
                                           *[None]*7)


class SimpleFunctionsTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           return_value=dict(frames=(3,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = SubFrameVDSGeneratorTester(
            files=["stripe_1.h5", "stripe_2.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])
        self.assertEqual(expected_source, source)

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           side_effect=[dict(frames=3, height=256, width=2048, dtype="uint16"),
                        dict(frames=4, height=256, width=2048,
                             dtype="uint16")])
    def test_process_source_datasets_given_mismatched_data(self, grab_mock):
        gen = SubFrameVDSGeneratorTester(
            files=["stripe_1.h5", "stripe_2.h5"])

        with self.assertRaises(ValueError):
            gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])

    def test_construct_vds_spacing(self):
        gen = SubFrameVDSGeneratorTester(
            files=[""] * 6, stripe_spacing=10, module_spacing=100)
        expected_spacing = [10, 100, 10, 100, 10, 0]

        spacing = gen.construct_vds_spacing()

        self.assertEqual(expected_spacing, spacing)

    file_mock = MagicMock()

    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(h5py_patch_path + '.VirtualSource')
    @patch(h5py_patch_path + '.VirtualLayout')
    def test_create_virtual_layout(self, layout_mock, source_mock, file_mock):
        gen = SubFrameVDSGeneratorTester(
            output_file="/test/path/vds.hdf5",
            stripe_spacing=10, module_spacing=100,
            target_node="full_frame", source_node="data",
            files=["raw1.h5", "raw2.h5", "raw3.h5",
                   "raw4.h5", "raw5.h5", "raw6.h5"],
            name="vds.hdf5")
        source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")
        dataset_mock = MagicMock()
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.__getitem__.return_value = dataset_mock

        layout = gen.create_virtual_layout(source)

        layout_mock.assert_called_once_with((3, 1766, 2048), "uint16")
        source_mock.assert_has_calls(
            [call("raw{}.h5".format(x),
                  dtype="uint16", name="data", shape=(3, 256, 2048))
             for x in range(1, 7)])
        # TODO: Pass numpy arrays to check slicing



"""A class for generating virtual dataset frames from sub-frames."""

import h5py as h5

from .vdsgenerator import VDSGenerator, SourceMeta


class GapFillVDSGenerator(VDSGenerator):

    """A class to generate a Virtual Dataset with gaps added to the source."""

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 sub_width=256, sub_height=256, grid_x=8, grid_y=2,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            sub_width(int): Width of sub-section in pixels
            sub_height(int): Height of sub-section in pixels
            grid_x(int): Width of full sensor in sub-sections
            grid_y(int): Height of full sensor in sub-sections
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        self.sub_width = sub_width
        self.sub_height = sub_height
        self.grid_x = grid_x
        self.grid_y = grid_y

        super(GapFillVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

        if len(self.files) > 1:
            raise ValueError("Can only insert gaps with a single dataset")

        self.source_file = self.files[0]  # We only have one dataset

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.files[0])

        source = SourceMeta(frames=data['frames'],
                            height=data['height'], width=data['width'],
                            dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", *source)
        return source

    def construct_vds_spacing(self):
        """Construct list of spacings between each sub-section.

        Returns:
            tuple(list): A list of spacings for horizontal and vertical gaps

        """
        raise NotImplementedError("Must be implemented in child class")

    def create_virtual_layout(self, source_meta):
        """Create a VirtualLayout mapping raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            VirtualLayout: Object describing links between raw data and VDS

        """
        x_spacing, y_spacing = self.construct_vds_spacing()

        target_shape = source_meta.frames + \
            (source_meta.height + sum(y_spacing),
             source_meta.width + sum(x_spacing))
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n", target_shape)

        v_layout = h5.VirtualLayout(target_shape, source_meta.dtype)

        source_shape = source_meta.frames + \
            (source_meta.height, source_meta.width)
        v_source = h5.VirtualSource(
            self.source_file, name=self.source_node,
            shape=source_shape, dtype=source_meta.dtype
        )

        y_current = 0
        for row_idx in range(self.grid_y):
            y_start = y_current
            y_stop = y_start + self.sub_height
            y_current = y_stop + y_spacing[row_idx]

            source_y_start = self.sub_height * row_idx
            source_y_end = source_y_start + self.sub_height

            x_current = 0
            for column_idx in range(self.grid_x):
                x_start = x_current
                x_stop = x_start + self.sub_width
                x_current = x_stop + x_spacing[column_idx]

                source_x_start = self.sub_width * column_idx
                source_x_end = source_x_start + self.sub_width

                # Hyperslab: All frames,
                #            Height bounds of chip,
                #            Width bounds of chip
                source_hyperslab = \
                    v_source[
                        :,
                        source_y_start:source_y_end,
                        source_x_start:source_x_end
                    ]

                # Hyperslab: All frames,
                #            Height bounds of chip with cumulative gap offset,
                #            Width bounds of chip with cumulative gap offset
                v_layout[:, y_start:y_stop, x_start:x_stop] = source_hyperslab

                self.logger.debug(
                    "Mapping %s[:, %s:%s, %s:%s] to %s[:, %s:%s, %s:%s].",
                    self.name,
                    source_y_start, source_y_end, source_x_start, source_x_end,
                    self.source_file.split("/")[-1],
                    y_start, y_stop, x_start, x_stop)

        return v_layout



import sys
import logging
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter,\
    RawTextHelpFormatter

from .vdsgenerator import VDSGenerator
from .interleavevdsgenerator import InterleaveVDSGenerator
from .subframevdsgenerator import SubFrameVDSGenerator
from .excaliburgapfillvdsgenerator import ExcaliburGapFillVDSGenerator
from .reshapevdsgenerator import ReshapeVDSGenerator

help_message = """
A script to create a virtual dataset composed of multiple raw HDF5 files.
The minimum required arguments are <path> and either -p <prefix> or -f <files>.

For example:
 > ../vdsgen/app.py /scratch/images -p stripe_
 > ../vdsgen/app.py /scratch/images -f stripe_1.hdf5 stripe_2.hdf5
 
You can create an empty VDS, for raw files that don't exist yet, with the -e
flag; you will then need to provide --shape and --data_type, though defaults
are provided for these.
"""


class Formatter(ArgumentDefaultsHelpFormatter, RawTextHelpFormatter):
    pass


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(description=help_message,
                            formatter_class=Formatter)
    parser.add_argument(
        "path", type=str, help="Root folder of source files and VDS.")

    # Definition of file names in <path> - Common prefix or explicit list
    file_definition = parser.add_mutually_exclusive_group(required=True)
    file_definition.add_argument(
        "-p", "--prefix", type=str, default=None, dest="prefix",
        help="Prefix of files to search <path> for - e.g 'stripe_' to combine "
             "'stripe_1.hdf5' and 'stripe_2.hdf5'.")
    file_definition.add_argument(
        "-f", "--files", nargs="*", type=str, default=None, dest="files",
        help="Explicit names of raw files in <path>.")

    # Arguments required to allow VDS to be created before raw files exist
    empty_vds = parser.add_argument_group(
        "Arguments required to describe the source dataset(s) when --empty is "
        "used"
    )
    empty_vds.add_argument(
        "-e", "--empty", action="store_true", dest="empty",
        help="Make empty VDS pointing to datasets that don't exist yet.")
    empty_vds.add_argument(
        "--shape", type=int, nargs="*", default=[1, 256, 2048], dest="shape",
        help="Shape of dataset - '<frames> <height> <width>', "
             "where frames is N dimensional.")
    empty_vds.add_argument(
        "-t", "--data-type", type=str, default="uint16", dest="data_type",
        help="Data type of raw datasets.")

    # Arguments related to the mode used
    mode_args = parser.add_argument_group(
        "Arguments that only take effect in certain modes"
    )
    mode_args.add_argument(
        "--mode", type=str, dest="mode", default="sub-frames",
        choices=["sub-frames", "interleave", "gap-fill", "reshape"],
        help="Type of VDS to create and expected input dataset(s)\n"
             "  sub-frames:    ND datasets containing sub-frames of full "
             "images\n"
             "  interleave:    1D datasets containing interspersed blocks of "
             "frames\n"
             "  gap-fill:      Single ND raw full-frame excalibur dataset to "
             "insert gaps into\n"
             "  reshape:       Single 1D dataset to reshape\n")
    mode_args.add_argument(
        "-s", "--stripe-spacing", type=int, dest="stripe_spacing", default=3,
        help="Spacing between two stripes in an Excalibur module. "
             "[gap-fill, sub-frames]")
    mode_args.add_argument(
        "-m", "--module-spacing", type=int, dest="module_spacing", default=123,
        help="Spacing between Excalibur modules. [gap-fill, sub-frames]")
    mode_args.add_argument(
        "-M", "--modules", type=int, dest="modules", choices=[1, 3], default=1,
        help="Number of modules in Excalibur sensor (1[M] or 3[M]). "
             "[gap-fill]")
    mode_args.add_argument(
        "-b", "--block-size", type=int, dest="block_size", default=1,
        help="Size of blocks of contiguous frames. [interleave]")
    mode_args.add_argument(
        "-S", "--new-shape", type=int, dest="new_shape", nargs="*",
        help="Shape to map 1D dataset into. [reshape]")
    mode_args.add_argument(
        "-A", "--alternate", type=bool, dest="alternate", nargs="*",
        help="Whether each axis alternates. List of True/False for each axis. "
             "[reshape]")

    # Arguments that always apply
    other_args = parser.add_argument_group()
    other_args.add_argument(
        "-o", "--output", type=str, default=None, dest="output",
        help="Output file name. If None then generated as input file prefix "
             "with vds suffix.")
    other_args.add_argument(
        "-F", "--fill-value", type=int, dest="fill_value", default=0,
        help="Fill value for missing data and gaps.")
    other_args.add_argument(
        "--source-node", type=str, dest="source_node",
        default=VDSGenerator.source_node,
        help="Data node in source HDF5 files.")
    other_args.add_argument(
        "--target-node", type=str, dest="target_node",
        default=VDSGenerator.target_node, help="Data node in VDS file.")
    other_args.add_argument(
        "-l", "--log-level", type=int, dest="log_level", choices=[1, 2, 3],
        default=VDSGenerator.log_level,
        help="Logging level (off=3, info=2, debug=1).")

    args = parser.parse_args()
    args.shape = tuple(args.shape)

    if args.empty and args.files is None:
        parser.error(
            "To make an empty VDS you must explicitly define --files for the "
            "eventual raw datasets."
        )
    # Check correct number of files given for the selected mode
    if args.files is not None:
        if args.mode == "gap-fill" and len(args.files) != 1:
            parser.error("Gap fill can only operate on a single dataset.")
    if args.mode == "reshape" and args.new_shape is None:
        parser.error("Must provide --new-shape for reshape mode")

    return args


def main():
    """Run program."""
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    args = parse_args()

    if args.empty:
        source_metadata = dict(shape=args.shape, dtype=args.data_type)
    else:
        source_metadata = None

    if args.mode == "interleave":
        gen = InterleaveVDSGenerator(
            args.path,
            prefix=args.prefix, files=args.files,
            output=args.output,
            source=source_metadata,
            source_node=args.source_node,
            target_node=args.target_node,
            block_size=args.block_size,
            fill_value=args.fill_value,
            log_level=args.log_level)
    elif args.mode == "sub-frames":
        gen = SubFrameVDSGenerator(
            args.path,
            prefix=args.prefix, files=args.files,
            output=args.output,
            source=source_metadata,
            source_node=args.source_node,
            target_node=args.target_node,
            stripe_spacing=args.stripe_spacing,
            module_spacing=args.module_spacing,
            fill_value=args.fill_value,
            log_level=args.log_level)
    elif args.mode == "gap-fill":
        gen = ExcaliburGapFillVDSGenerator(
            args.path,
            prefix=args.prefix, files=args.files,
            output=args.output,
            source=source_metadata,
            source_node=args.source_node,
            target_node=args.target_node,
            modules=args.modules,
            chip_spacing=args.stripe_spacing,
            module_spacing=args.module_spacing,
            fill_value=args.fill_value,
            log_level=args.log_level
        )
    elif args.mode == "reshape":
        gen = ReshapeVDSGenerator(
            tuple(args.new_shape),
            args.path,
            prefix=args.prefix, files=args.files,
            output=args.output,
            source=source_metadata,
            source_node=args.source_node,
            target_node=args.target_node,
            fill_value=args.fill_value,
            log_level=args.log_level,
            alternate=args.alternate
        )
    else:
        raise NotImplementedError("Invalid VDS mode. Must be frames, "
                                  "interleave, sub-frames, gap-fill "
                                  "or reshape.")

    gen.generate_vds()


if __name__ == "__main__":
    sys.exit(main())



"""A class for generating virtual dataset frames from sub-frames."""

import h5py as h5

from .vdsgenerator import VDSGenerator, SourceMeta


class SubFrameVDSGenerator(VDSGenerator):

    """A class to generate Virtual Dataset frames from sub-frames."""

    # Default Values
    stripe_spacing = 10  # Pixel spacing between stripes in a module
    module_spacing = 10  # Pixel spacing between modules

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 stripe_spacing=None, module_spacing=None,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            stripe_spacing(int): Spacing between stripes in module
            module_spacing(int): Spacing between modules
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        super(SubFrameVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

        # Overwrite default values with arguments, if given
        if stripe_spacing is not None:
            self.stripe_spacing = stripe_spacing
        if module_spacing is not None:
            self.module_spacing = module_spacing

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.files[0])
        frames = [data["frames"]]
        for dataset in self.files[1:]:
            temp_data = self.grab_metadata(dataset)
            frames.append(temp_data["frames"])
            for attribute, value in data.items():
                if temp_data[attribute] != value:
                    raise ValueError("Files have mismatched "
                                     "{}".format(attribute))

        source = SourceMeta(frames=data['frames'], height=data['height'],
                            width=data['width'], dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", frames, *source[1:])
        return source

    def construct_vds_spacing(self):
        """Construct list of spacing between each stripe.

        Returns:
            VDS: Dataset spacing of virtual data set

        """
        stripes = len(self.files)
        spacing = [0] * stripes
        for idx in range(0, stripes - 1, 2):
            spacing[idx] = self.stripe_spacing
        for idx in range(1, stripes, 2):
            spacing[idx] = self.module_spacing
        # We don't want the final stripe to have a gap afterwards
        spacing[-1] = 0

        return spacing

    def create_virtual_layout(self, source_meta):
        """Create a list of VirtualMaps of raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            list(VirtualMap): Maps describing links between raw data and VDS

        """
        source_shape = source_meta.frames + \
            (source_meta.height, source_meta.width)
        spacing = self.construct_vds_spacing()
        target_height = source_meta.height * len(self.files) + sum(spacing)
        target_shape = source_meta.frames + (target_height,
                                             source_meta.width)
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n"
                          "  Spacing: %s", target_shape, spacing)

        v_layout = h5.VirtualLayout(target_shape, source_meta.dtype)

        current_position = 0
        for stripe_idx, file_path in enumerate(self.files):
            v_source = h5.VirtualSource(
                file_path, name=self.source_node,
                shape=source_shape, dtype=source_meta.dtype
            )

            start = current_position
            end = start + source_meta.height
            current_position = end + spacing[stripe_idx]

            # Hyperslab: All frames for each axis,
            #            Height bounds of stripe,
            #            Entire width
            v_layout[..., start:end, :] = v_source

            self.logger.debug("Mapping %s[..., %s:%s, :] to %s[...].",
                              self.name, start, end,
                              file_path.split("/")[-1])

        return v_layout



import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.reshapevdsgenerator import ReshapeVDSGenerator

vdsgen_patch_path = "vdsgen.reshapevdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class ReshapeVDSGeneratorTester(ReshapeVDSGenerator):
    """A version of VDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)


class ReshapeVDSGeneratorInitTest(unittest.TestCase):

    super_mock = MagicMock()

    def set_files(self, _path, _prefix, files, *_):
        """A method to patch VDSGenerator __init__ with"""
        # Set self.files so the GapFill __init__ works
        self.files = files
        # Make a call we can test for
        ReshapeVDSGeneratorInitTest.super_mock(_path, _prefix, files, *_)

    @classmethod
    def setUp(cls):
        cls.super_mock.reset_mock()

    @patch(VDSGenerator_patch_path + ".__init__", new=set_files)
    def test_super_called(self):
        ReshapeVDSGenerator((3, 5, 10), "/test/path", files=["raw.h5"])

        self.super_mock.assert_called_once_with("/test/path", None, ["raw.h5"],
                                                *[None]*6)


class SimpleFunctionsTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + ".grab_metadata",
           return_value=dict(frames=(5,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = ReshapeVDSGeneratorTester(
            files=["raw.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(5,), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_called_once_with("raw.h5")
        self.assertEqual(expected_source, source)

    @patch(VDSGenerator_patch_path + ".grab_metadata",
           side_effect=[
               dict(frames=(3,), height=256, width=2048, dtype="uint16"),
               dict(frames=(3,), height=512, width=2048, dtype="uint16")])
    def test_process_source_datasets_given_mismatched_data(self, grab_mock):
        gen = ReshapeVDSGeneratorTester(
            files=["stripe_1.h5", "stripe_2.h5"])

        with self.assertRaises(ValueError):
            gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])

    file_mock = MagicMock()

    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(h5py_patch_path + '.VirtualSource')
    @patch(h5py_patch_path + '.VirtualLayout')
    def test_create_virtual_layout(self, layout_mock, source_mock, file_mock):
        gen = ReshapeVDSGeneratorTester(
            output_file="/test/path/vds.hdf5",
            dimensions=(5, 3, 10), alternate=None, periods=[], radices=(30, 10),
            target_node="full_frame", source_node="data",
            source_file="raw.h5", name="vds.hdf5")
        source = vdsgenerator.SourceMeta(
            frames=(150,), height=256, width=2048, dtype="uint16")
        dataset_mock = MagicMock()
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.__getitem__.return_value = dataset_mock

        layout = gen.create_virtual_layout(source)

        layout_mock.assert_called_once_with((5, 3, 10, 256, 2048), "uint16")
        source_mock.assert_called_once_with(
            "raw.h5", dtype="uint16", name="data", shape=(150, 256, 2048)
        )

    def test_create_virtual_layout_frame_mismatch(self):
        gen = ReshapeVDSGeneratorTester(
            output_file="/test/path/vds.hdf5",
            dimensions=(5, 3, 10), alternate=None, periods=[], radices=(30, 10),
            target_node="full_frame", source_node="data",
            source_file="raw.h5", name="vds.hdf5")
        source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        with self.assertRaises(ValueError):
            gen.create_virtual_layout(source)



"""A class for generating virtual datasets from individual HDF5 files."""

import os
import re
import logging

from collections import namedtuple

import h5py as h5

SourceMeta = namedtuple("SourceMeta", ["frames", "height", "width", "dtype"])


class VDSGenerator(object):

    """A class to generate Virtual Datasets from raw HDF5 files."""

    # Constants
    CREATE = "w"  # Will overwrite any existing file
    APPEND = "a"
    READ = "r"
    FULL_SLICE = slice(None)

    # Default Values
    fill_value = -1  # Fill value for spacing
    source_node = "data"  # Data node in source HDF5 files
    target_node = "data"  # Data node in VDS file
    mode = CREATE  # Write mode for vds file
    log_level = 2

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.log_level * 10)

        if (prefix is None and files is None) or \
                (prefix is not None and files is not None):
            raise ValueError("One, and only one, of prefix or files required.")

        self.path = path

        # Overwrite default values with arguments, if given
        if source_node is not None:
            self.source_node = source_node
        if target_node is not None:
            self.target_node = target_node
        if fill_value is not None:
            self.fill_value = fill_value
        if log_level is not None:
            self.logger.setLevel(log_level * 10)

        # If Files not given, find files using path and prefix.
        if files is None:
            self.prefix = prefix
            self.files = self.find_files()
            files = [path_.split("/")[-1] for path_ in self.files]
        # Else, get common prefix of given files and store full path
        else:
            self.prefix = os.path.commonprefix(files)
            self.files = [os.path.join(path, file_) for file_ in files]

        # If output vds file name given, use, otherwise generate a default
        if output is None:
            self.name = self.construct_vds_name(files)
        else:
            self.name = output

        # If source not given, check files exist and get metadata.
        if source is None:
            for file_ in self.files:
                if not os.path.isfile(file_):
                    raise IOError(
                        "File {} does not exist. To create VDS from raw "
                        "files that haven't been created yet, source "
                        "must be provided.".format(file_))
            self.source_metadata = self.process_source_datasets()
        # Else, store given source metadata
        else:
            self.source_metadata = self.process_source_metadata(source)

        self.output_file = os.path.abspath(os.path.join(self.path, self.name))

    def process_source_metadata(self, source):
        frames, height, width = self.parse_shape(source['shape'])
        source_metadata = SourceMeta(
            frames=frames, height=height, width=width,
            dtype=source['dtype'])

        return source_metadata

    @staticmethod
    def parse_shape(shape):
        """Split shape into height, width and frames.

        Args:
            shape(tuple): Shape of dataset

        Returns:
            frames, height, width

        """
        # The last two elements of shape are the height and width of the image
        height, width = shape[-2:]
        # Everything before that is the frames for each axis
        frames = shape[:-2]

        return frames, height, width

    def generate_vds(self):
        """Generate a virtual dataset."""
        if os.path.isfile(self.output_file):
            with h5.File(self.output_file, self.READ, libver="latest") as vds:
                node = vds.get(self.target_node)
            if node is not None:
                raise IOError("VDS {file} already has an entry for node "
                              "{node}".format(file=self.output_file,
                                              node=self.target_node))
            else:
                self.mode = self.APPEND

        virtual_layout = self.create_virtual_layout(self.source_metadata)

        self.logger.info("Creating VDS at %s", self.output_file)
        with h5.File(self.output_file, self.mode, libver="latest") as vds:
            self.validate_node(vds)
            vds.create_virtual_dataset(self.target_node, virtual_layout,
                                       fillvalue=self.fill_value)

    def find_files(self):
        """Find HDF5 files in given folder with given prefix.

        Returns:
            list: HDF5 files in folder that have the given prefix

        """
        regex = re.compile(self.prefix + r"\d+\.(hdf5|hdf|h5)")

        files = []
        for file_ in sorted(os.listdir(self.path)):
            if re.match(regex, file_):
                files.append(os.path.abspath(os.path.join(self.path, file_)))

        if len(files) == 0:
            raise IOError("No files matching pattern found. Got path: {path}, "
                          "prefix: {prefix}".format(path=self.path,
                                                    prefix=self.prefix))
        elif len(files) < 2:
            raise IOError("Folder must contain more than one matching HDF5 "
                          "file.")
        else:
            self.logger.debug("Found datasets:\n  %s",
                              ", ".join([f.split("/")[-1] for f in files]))
            return files

    def construct_vds_name(self, files):
        """Generate the file name for the VDS from the sub files.

        Args:
            files(list(str)): HDF5 files being combined

        Returns:
            str: Name of VDS file

        """
        _, ext = os.path.splitext(files[0])
        vds_name = "{prefix}vds{ext}".format(prefix=self.prefix, ext=ext)

        self.logger.debug("Generated VDS name:\n  %s", vds_name)
        return vds_name

    def grab_metadata(self, file_path):
        """Grab data from given HDF5 file.

        Args:
            file_path(str): Path to HDF5 file

        Returns:
            dict: Number of frames, height, width and data type of datasets

        """
        h5_data = h5.File(file_path, 'r')[self.source_node]
        frames, height, width = self.parse_shape(h5_data.shape)
        data_type = h5_data.dtype

        return dict(frames=frames, height=height, width=width, dtype=data_type)

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        raise NotImplementedError("Must be implemented in child class")

    def create_virtual_layout(self, source_meta):
        """Create a VirtualLayout mapping raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            VirtualLayout: Object describing links between raw data and VDS

        """
        raise NotImplementedError("Must be implemented in child class")

    def validate_node(self, vds_file):
        """Check if it is possible to create the given node.

        Create any sub-group of the target node if it doesn't exist.

        Args:
            vds_file(h5py.File): File to check for node

        """
        while self.target_node.endswith("/"):
            self.target_node = self.target_node[:-1]

        if "/" in self.target_node:
            sub_group = self.target_node.rsplit("/", 1)[0]
            if vds_file.get(sub_group) is None:
                vds_file.create_group(sub_group)



"""A class for generating virtual dataset frames from sub-frames."""

import h5py as h5

from .vdsgenerator import VDSGenerator, SourceMeta


class InterleaveVDSGenerator(VDSGenerator):

    """A class to generate Virtual Dataset frames from sub-frames."""

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 block_size=1,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            block_size(int): Number of contiguous frames per block
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        self.block_size = block_size

        super(InterleaveVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.files[0])
        frames = [data["frames"][0]]
        for dataset in self.files[1:]:
            temp_data = self.grab_metadata(dataset)
            frames.append(temp_data["frames"][0])
            for attribute, value in data.items():
                if attribute != "frames" and temp_data[attribute] != value:
                    raise ValueError("Files have mismatched "
                                     "{}".format(attribute))

        source = SourceMeta(frames=tuple(frames),
                            height=data['height'], width=data['width'],
                            dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", frames, *source[1:])
        return source

    def process_source_metadata(self, source):
        frames, height, width = self.parse_shape(source["shape"])
        if not isinstance(frames[0], tuple):
            raise ValueError(
                "For InterleaveVDSGenerator, source.shape.frames must be a "
                "tuple of the number of frames in each raw file."
            )
        source_metadata = SourceMeta(
            frames=frames[0], height=height, width=width,
            dtype=source["dtype"])

        return source_metadata

    def create_virtual_layout(self, source_meta):
        """Create a VirtualLayout mapping raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            VirtualLayout: Object describing links between raw data and VDS

        """
        total_frames = sum(source_meta.frames)
        target_shape = (total_frames,) + \
                       (source_meta.height, source_meta.width)
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n", target_shape)

        v_layout = h5.VirtualLayout(target_shape, source_meta.dtype)

        total_files = len(self.files)
        for file_idx, file_path in enumerate(self.files):
            source_shape = (source_meta.frames[file_idx],) + \
                (source_meta.height, source_meta.width)
            v_source = h5.VirtualSource(
                file_path, name=self.source_node,
                shape=source_shape, dtype=source_meta.dtype
            )
            dataset_frames = v_source.shape[0]

            for frame_idx in range(0, dataset_frames, self.block_size):

                source_block_idx = frame_idx // self.block_size
                target_block_idx = \
                    file_idx + total_files * source_block_idx

                # Make sure to only take as many frames as raw dataset has
                # It is OK if this isn't a complete block if it is the last one
                block_size = min(self.block_size, dataset_frames - frame_idx)

                if block_size < self.block_size:
                    self.logger.warning(
                        "%s does not have integer number of blocks sized %s",
                        file_path.split("/")[-1], self.block_size)

                source_start = frame_idx
                source_end = frame_idx + block_size
                # Hyperslab: Frame[block_start_idx, block_end_idx + 1],
                #            Full slice for height and width
                source_hyperslab = v_source[source_start: source_end, :, :]

                target_start = self.block_size * target_block_idx
                if target_start + block_size > total_frames:
                    raise RuntimeError(
                        "Cannot map dataset %d of %d [%d frames] "
                        "to vds [%d frames] with blocks sized %d" %
                        (self.files.index(file_path) + 1, len(self.files),
                         dataset_frames, total_frames, self.block_size))
                target_end = target_start + block_size

                # Hyperslab: Frame[block_start_idx, block_end_idx + 1],
                #            Full slice for height and width
                v_layout[target_start:target_end, :, :] = source_hyperslab

                self.logger.debug(
                    "Mapping %s[%s:%s, :, :] to %s[%s:%s, :, :].",
                    self.name, target_start, target_end,
                    file_path.split("/")[-1], source_start, source_end)

        return v_layout



from setuptools import setup, find_packages

MODULE_NAME = "vdsgen"

setup(
    name=MODULE_NAME,
    version='0.4',
    description='Creates virtual dataset HDF5 files',
    long_description=open("README.rst").read(),
    url='https://github.com/dls-controls/vds-gen',
    author='Gary Yendell',
    author_email='gary.yendell@diamond.ac.uk',
    keywords='',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
    ],
    license='APACHE',
    include_package_data=True,
    test_suite='nose.collector',
    install_requires=[
        'h5py==2.9.0'
    ],
    tests_require=[
        'nose',
        'mock'
    ],
    zip_safe=False,
    entry_points={'console_scripts': ["dls-vds-gen.py = vdsgen.app:main"]}
)



import os
import sys
import unittest
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.vdsgenerator import VDSGenerator

vdsgen_patch_path = "vdsgen.vdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class VDSGeneratorTester(VDSGenerator):

    """A version of VDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class VDSGeneratorInitTest(unittest.TestCase):

    @patch('os.path.isfile', return_value=True)
    @patch(VDSGenerator_patch_path + '.process_source_datasets')
    @patch(VDSGenerator_patch_path + '.construct_vds_name',
           return_value="stripe_vds.hdf5")
    @patch(VDSGenerator_patch_path + '.find_files',
           return_value=["/test/path/stripe_1.hdf5",
                         "/test/path/stripe_2.hdf5",
                         "/test/path/stripe_3.hdf5"])
    def test_generate_vds_defaults(self, find_mock, construct_mock,
                                   process_mock, isfile_mock):
        expected_files = ["stripe_1.hdf5", "stripe_2.hdf5", "stripe_3.hdf5"]

        gen = VDSGenerator("/test/path", prefix="stripe_")

        find_mock.assert_called_once_with()
        construct_mock.assert_called_once_with(expected_files)
        process_mock.assert_called_once_with()

        self.assertEqual("/test/path", gen.path)
        self.assertEqual("stripe_", gen.prefix)
        self.assertEqual("stripe_vds.hdf5", gen.name)
        self.assertEqual(find_mock.return_value, gen.files)
        self.assertEqual(process_mock.return_value, gen.source_metadata)
        self.assertEqual("data", gen.source_node)
        self.assertEqual("data", gen.target_node)
        self.assertEqual(gen.CREATE, gen.mode)

    def test_generate_vds_given_args(self):
        files = ["stripe_1.h5", "stripe_2.h5"]
        file_paths = ["/test/path/" + file_ for file_ in files]
        source_dict = dict(shape=(3, 256, 2048), dtype="int16")
        source = vdsgenerator.SourceMeta(frames=(3,), height=256, width=2048,
                                     dtype="int16")

        gen = VDSGenerator("/test/path",
                           files=files,
                           output="vds.hdf5",
                           source=source_dict,
                           source_node="entry/data/data",
                           target_node="entry/detector/detector1")

        self.assertEqual("/test/path", gen.path)
        self.assertEqual("stripe_", gen.prefix)
        self.assertEqual("vds.hdf5", gen.name)
        self.assertEqual(file_paths, gen.files)
        self.assertEqual(source, gen.source_metadata)
        self.assertEqual("entry/data/data", gen.source_node)
        self.assertEqual("entry/detector/detector1", gen.target_node)
        self.assertEqual(gen.CREATE, gen.mode)

    def test_generate_vds_prefix_and_files_then_error(self):
        files = ["stripe_1.h5", "stripe_2.h5"]
        source_dict = dict(frames=3, height=256, width=2048, dtype="int16")

        with self.assertRaises(ValueError):
            VDSGenerator("/test/path",
                         prefix="stripe_", files=files,
                         output="vds.hdf5",
                         source=source_dict,
                         source_node="entry/data/data",
                         target_node="entry/detector/detector1")

    @patch('os.path.isfile', return_value=False)
    def test_generate_vds_no_source_or_files_then_error(self, _):

        with self.assertRaises(IOError) as e:
            VDSGenerator("/test/path",
                         files=["file1", "file2"],
                         output="vds.hdf5")

        self.assertEqual("File /test/path/file1 does not exist. To create VDS "
                         "from raw files that haven't been created yet, "
                         "source must be provided.",
                         str(e.exception))


class FindFilesTest(unittest.TestCase):

    def setUp(self):
        self.gen = VDSGeneratorTester(path="/test/path", prefix="stripe_")

    @patch('os.listdir',
           return_value=["stripe_1.h5", "stripe_2.h5", "stripe_3.h5",
                         "stripe_4.h5", "stripe_5.h5", "stripe_6.h5"])
    def test_given_files_then_return(self, _):
        expected_files = ["/test/path/stripe_1.h5", "/test/path/stripe_2.h5",
                          "/test/path/stripe_3.h5", "/test/path/stripe_4.h5",
                          "/test/path/stripe_5.h5", "/test/path/stripe_6.h5"]

        files = self.gen.find_files()

        self.assertEqual(expected_files, files)

    @patch('os.listdir',
           return_value=["stripe_4.h5", "stripe_1.h5", "stripe_6.h5",
                         "stripe_3.h5", "stripe_2.h5", "stripe_5.h5"])
    def test_given_files_out_of_order_then_return(self, _):
        expected_files = ["/test/path/stripe_1.h5", "/test/path/stripe_2.h5",
                          "/test/path/stripe_3.h5", "/test/path/stripe_4.h5",
                          "/test/path/stripe_5.h5", "/test/path/stripe_6.h5"]

        files = self.gen.find_files()

        self.assertEqual(expected_files, files)

    @patch('os.listdir', return_value=["stripe_1.h5"])
    def test_given_one_file_then_error(self, _):

        with self.assertRaises(IOError):
            self.gen.find_files()

    @patch('os.listdir', return_value=[])
    def test_given_no_files_then_error(self, _):

        with self.assertRaises(IOError):
            self.gen.find_files()


class SimpleFunctionsTest(unittest.TestCase):

    def test_generate_vds_name(self):
        gen = VDSGeneratorTester(prefix="stripe_")
        expected_name = "stripe_vds.h5"
        files = ["stripe_1.h5", "stripe_2.h5", "stripe_3.h5",
                 "stripe_4.h5", "stripe_5.h5", "stripe_6.h5"]

        vds_name = gen.construct_vds_name(files)

        self.assertEqual(expected_name, vds_name)

    mock_data = dict(data=MagicMock(shape=(3, 256, 2048), dtype="uint16"))

    @patch(h5py_patch_path + '.File', return_value=mock_data)
    def test_grab_metadata(self, h5file_mock):
        gen = VDSGeneratorTester(source_node="data")
        expected_data = dict(frames=(3,), height=256, width=2048, dtype="uint16")

        meta_data = gen.grab_metadata("/test/path/stripe.hdf5")

        h5file_mock.assert_called_once_with("/test/path/stripe.hdf5", "r")
        self.assertEqual(expected_data, meta_data)

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           side_effect=[dict(frames=3, height=256, width=2048, dtype="uint16"),
                        dict(frames=4, height=256, width=2048,
                             dtype="uint16")])
    def test_process_source_datasets_given_mismatched_data(self, grab_mock):
        gen = VDSGeneratorTester(datasets=["stripe_1.h5", "stripe_2.h5"])

        with self.assertRaises(NotImplementedError):
            gen.process_source_datasets()

    def test_create_virtual_layout(self):
        gen = VDSGeneratorTester(output_file="/test/path/vds.hdf5",
                                 stripe_spacing=10, module_spacing=100,
                                 target_node="full_frame", source_node="data",
                                 datasets=["source"] * 6, name="vds.hdf5")
        source = vdsgenerator.SourceMeta(frames=(3,), height=256, width=2048,
                                         dtype="uint16")

        with self.assertRaises(NotImplementedError):
            gen.create_virtual_layout(source)


class ValidateNodeTest(unittest.TestCase):

    def setUp(self):
        self.file_mock = MagicMock()

    def test_validate_node_creates(self):
        gen = VDSGeneratorTester(target_node="/entry/detector/detector1")
        self.file_mock.get.return_value = None

        gen.validate_node(self.file_mock)

        self.file_mock.create_group.assert_called_once_with("/entry/detector")

    def test_validate_node_exists_then_no_op(self):
        gen = VDSGeneratorTester(target_node="entry/detector/detector1")
        self.file_mock.get.return_value = "Group"

        gen.validate_node(self.file_mock)

        self.file_mock.create_group.assert_not_called()

    def test_validate_node_trailing_slash_then_removed(self):
        gen = VDSGeneratorTester(target_node="/entry/detector/detector1//")
        self.file_mock.get.return_value = None

        gen.validate_node(self.file_mock)

        self.file_mock.create_group.assert_called_once_with("/entry/detector")


class GenerateVDSTest(unittest.TestCase):

    file_mock = MagicMock()

    @patch('os.path.isfile', return_value=False)
    @patch(VDSGenerator_patch_path + '.validate_node')
    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(VDSGenerator_patch_path + '.create_virtual_layout')
    def test_generate_vds_create(self, create_mock,
                                 h5file_mock, validate_mock, isfile_mock):
        source_mock = MagicMock()
        gen = VDSGeneratorTester(path="/test/path", prefix="stripe_",
                                 output_file="/test/path/vds.hdf5",
                                 name="vds.hdf5",
                                 target_node="full_frame", source_node="data",
                                 datasets=["stripe_1.hdf5", "stripe_2.hdf5",
                                           "stripe_3.hdf5"],
                                 source_metadata=source_mock)
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.get.return_value = None

        gen.generate_vds()

        isfile_mock.assert_called_once_with("/test/path/vds.hdf5")
        create_mock.assert_called_once_with(source_mock)
        validate_mock.assert_called_once_with(vds_file_mock)
        h5file_mock.assert_called_once_with(
            "/test/path/vds.hdf5", "w", libver="latest")
        vds_file_mock.create_virtual_dataset.assert_called_once_with(
            "full_frame", create_mock.return_value, fillvalue=-1)

    @patch('os.path.isfile', return_value=True)
    @patch(VDSGenerator_patch_path + '.validate_node')
    @patch(h5py_patch_path + '.File', return_value=file_mock)
    @patch(VDSGenerator_patch_path + '.create_virtual_layout')
    def test_generate_vds_append(self, create_mock,
                                 h5file_mock, validate_mock, isfile_mock):
        source_mock = MagicMock()
        gen = VDSGeneratorTester(path="/test/path", prefix="stripe_",
                                 output_file="/test/path/vds.hdf5",
                                 name="vds.hdf5",
                                 target_node="full_frame", source_node="data",
                                 datasets=["stripe_1.hdf5", "stripe_2.hdf5",
                                           "stripe_3.hdf5"],
                                 source_metadata=source_mock)
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.get.return_value = None

        gen.generate_vds()

        isfile_mock.assert_called_once_with("/test/path/vds.hdf5")
        create_mock.assert_called_once_with(source_mock)
        validate_mock.assert_called_once_with(vds_file_mock)
        h5file_mock.assert_has_calls([
            call("/test/path/vds.hdf5", "r", libver="latest"),
            call("/test/path/vds.hdf5", "a", libver="latest")])
        vds_file_mock.create_virtual_dataset.assert_called_once_with(
            "full_frame", create_mock.return_value, fillvalue=-1)

    @patch('os.path.isfile', return_value=True)
    @patch(h5py_patch_path + '.File', return_value=file_mock)
    def test_generate_vds_node_exists_then_error(self, h5file_mock,
                                                 isfile_mock):
        source_mock = MagicMock()
        gen = VDSGeneratorTester(path="/test/path", prefix="stripe_",
                                 output_file="/test/path/vds.hdf5",
                                 name="vds.hdf5",
                                 target_node="full_frame", source_node="data",
                                 datasets=["stripe_1.hdf5", "stripe_2.hdf5",
                                           "stripe_3.hdf5"],
                                 source_metadata=source_mock)
        self.file_mock.reset_mock()
        vds_file_mock = self.file_mock.__enter__.return_value
        vds_file_mock.get.return_value = "Group"

        with self.assertRaises(IOError):
            gen.generate_vds()



"""A class to generate a Virtual Dataset with gaps added to the source."""

from .gapfillvdsgenerator import GapFillVDSGenerator


class ExcaliburGapFillVDSGenerator(GapFillVDSGenerator):

    """A class to generate a Virtual Dataset with gaps added to the source."""

    GRID_X = 8  # All Excalibur sensors are 8 chips wide
    CHIP_SIZE = 256  # Width and height of Excalibur chips is 256 pixels

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 modules=1, chip_spacing=3, module_spacing=10,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            modules(int): Number of modules in sensor (1 or 3)
            chip_spacing(int): Spacing between stripes in module
            module_spacing(int): Spacing between modules
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        if modules == 1:
            grid_y = 2
        elif modules == 3:
            grid_y = 6
        else:
            raise ValueError("Invalid number of modules - Must be 1 or 3")

        self.chip_spacing = chip_spacing
        self.module_spacing = module_spacing

        super(ExcaliburGapFillVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value,
            self.CHIP_SIZE, self.CHIP_SIZE, self.GRID_X, grid_y,
            log_level)

    def construct_vds_spacing(self):
        """Construct lists of x and y spacings between sub-sections.

        Returns:
            tuple(list): A list of spacings for horizontal and vertical gaps

        """
        x_spacing = [self.chip_spacing] * self.GRID_X
        x_spacing[-1] = 0  # No gap on end

        y_spacing = [self.chip_spacing] * self.grid_y
        for idx in range(1, self.grid_y, 2):
            y_spacing[idx] = self.module_spacing
        y_spacing[-1] = 0  # No gap at bottom

        return x_spacing, y_spacing



from __future__ import print_function
import os
import sys
from timeit import default_timer as timer
import logging
from unittest import TestCase

import numpy as np
import h5py as h5

from vdsgen import SubFrameVDSGenerator, InterleaveVDSGenerator, \
    ExcaliburGapFillVDSGenerator, ReshapeVDSGenerator, generate_raw_files


def create_pixel_pattern(top_value, bottom_value,
                         horizontal_gap, vertical_gap,
                         fill_value=-1.0):
    WIDTH = 5
    pattern = [[top_value] * WIDTH,
               [bottom_value] * WIDTH]

    # Insert rows for horizontal gaps
    for _ in range(horizontal_gap):
        pattern.insert(1, [fill_value] * WIDTH)

    # Insert elements in all rows for vertical gaps
    if vertical_gap != 0:
        for row in pattern:
            row[1: WIDTH - 1] = [fill_value] * vertical_gap

    return np.array(pattern)


class SystemTest(TestCase):

    def setUp(self):
        logging.basicConfig(
            format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

        for file_ in os.listdir("./"):
            if file_.endswith(".h5"):
                os.remove(file_)

    def tearDown(self):
        for file_ in os.listdir("./"):
            if file_.endswith(".h5"):
                os.remove(file_)

    def test_interleave(self):
        FRAMES = 95
        WIDTH = 2048
        HEIGHT = 1536
        # Generate 4 raw files with interspersed frames
        # 95 2048x1536 frames, between 4 files in blocks of 10
        print("Creating raw files...")
        generate_raw_files("OD", FRAMES, 4, 10, WIDTH, HEIGHT, "test_data")
        print("Creating VDS...")
        gen = InterleaveVDSGenerator(
            "./", prefix="OD_", block_size=10, log_level=1,
            source_node="test_data"
        )
        gen.generate_vds()

        self._validate_interleave(FRAMES, HEIGHT, WIDTH)

    def test_interleave_empty(self):
        FRAMES = 95
        WIDTH = 2048
        HEIGHT = 1536
        print("Creating VDS...")
        gen = InterleaveVDSGenerator(
            "./", files=["OD_0.h5", "OD_1.h5", "OD_2.h5", "OD_3.h5"],
            source=dict(shape=((30, 25, 20, 20), HEIGHT, WIDTH),
                        dtype="float32"),
            block_size=10, log_level=1
        )
        gen.generate_vds()
        # Generate 4 raw files with interspersed frames
        # 95 2048x1536 frames, between 4 files in blocks of 10
        print("Creating raw files...")
        generate_raw_files("OD", FRAMES, 4, 10, WIDTH, HEIGHT)

        self._validate_interleave(FRAMES, HEIGHT, WIDTH)

    def _validate_interleave(self, frames, height, width):
        print("Opening VDS...")
        with h5.File("OD_vds.h5", mode="r") as h5_file:
            vds_dataset = h5_file["data"]

            print("Verifying dataset...")
            # Check shape
            self.assertEqual(vds_dataset.shape, (frames, height, width))
            # Check first pixel of each frame
            print("0 %", end="")
            for frame_idx in range(frames):
                self.assertEqual(vds_dataset[frame_idx][0][0],
                                 1.0 * frame_idx)
                progress = int((frame_idx + 1) * (1.0 / frames * 100))
                print("\r{} %".format(progress), end="")
                sys.stdout.flush()
            print()

    def test_interleave_speed(self):
        FILES = 4
        FRAMES = 10000
        WIDTH = 2048
        HEIGHT = 1536
        print("Creating VDS...")
        start = timer()
        gen = InterleaveVDSGenerator(
            "./", files=["OD_{}.h5".format(idx) for idx in range(FILES)],
            source=dict(shape=((FRAMES // FILES,) * FILES, HEIGHT, WIDTH),
                        dtype="float32"),
            block_size=10, log_level=1
        )
        gen.generate_vds()
        end = timer()
        print("Interleave {} frames - Time: {}".format(FRAMES, end - start))

    def test_sub_frames(self):
        FRAMES = 1
        WIDTH = 2048
        HEIGHT = 256
        FEMS = 6
        CHIPS = 8
        # Generate 6 raw files each with 1/6th of a single 2048x1536 frame
        print("Creating raw files...")
        generate_raw_files("stripe", FEMS * FRAMES, FEMS, 1, WIDTH, HEIGHT)
        print("Creating VDS...")
        gen = SubFrameVDSGenerator(
            "./", prefix="stripe_", stripe_spacing=3, module_spacing=123,
            log_level=1
        )
        gen.generate_vds()

        self._validate_sub_frames(FRAMES, WIDTH, FEMS, CHIPS)

    def test_sub_frames_empty(self):
        FRAMES = 1
        WIDTH = 2048
        HEIGHT = 256
        FEMS = 6
        CHIPS = 8
        print("Creating VDS...")
        gen = SubFrameVDSGenerator(
            "./", files=["stripe_0.h5", "stripe_1.h5", "stripe_2.h5",
                         "stripe_3.h5", "stripe_4.h5", "stripe_5.h5"],
            source=dict(shape=(FRAMES, HEIGHT, WIDTH), dtype="float32"),
            stripe_spacing=3, module_spacing=123,
            log_level=1
        )
        gen.generate_vds()
        # Generate 6 raw files each with 1/6th of a single 2048x1536 frame
        print("Creating raw files...")
        generate_raw_files("stripe", FEMS * FRAMES, FEMS, 1, WIDTH, HEIGHT)

        self._validate_sub_frames(FRAMES, WIDTH, FEMS, CHIPS)

    def _validate_sub_frames(self, frames, width, fems, chips):
        print("Opening VDS...")
        with h5.File("stripe_vds.h5", mode="r") as h5_file:
            vds_dataset = h5_file["data"]
            print("Verifying dataset...")
            # Check shape
            self.assertEqual(vds_dataset.shape, (frames, 1791, width))

            self.assertEqual(vds_dataset[0][0][0], 0.0)  # FEM 1 pixel 1

            # Check corners where 4 chips meet have expected pattern
            row = 255
            column = 255
            spacings = [3, 123, 3, 123, 3]
            frame = vds_dataset[0]
            for stripe_idx in range(fems - 1):  # FEMS - 1 Vertical Gaps
                spacing = spacings[stripe_idx]
                pattern = create_pixel_pattern(
                    top_value=1.0 * stripe_idx,
                    bottom_value=1.0 * (stripe_idx + 1),
                    horizontal_gap=spacing, vertical_gap=0, fill_value=-1.0
                )
                for chip_idx in range(chips - 1):  # CHIPS - 1 Horizontal Gaps
                    test = frame[row: row + spacing + 2, column: column + 5]
                    np.testing.assert_array_equal(pattern, test)
                    column += 256 + 3
                column = 255
                row += 256 + spacing

    def test_gap_fill(self):
        FEMS = 6
        CHIPS = 8
        FRAMES = 100
        HEIGHT = 1536
        WIDTH = 2048
        # Generate a single file with 100 2048x1536 frames
        print("Creating raw files...")
        generate_raw_files("raw", FRAMES, 1, 1, WIDTH, HEIGHT)
        print("Creating VDS...")
        gen = ExcaliburGapFillVDSGenerator(
            "./", files=["raw_0.h5"], chip_spacing=3, module_spacing=123,
            modules=3, output="gaps.h5", log_level=1
        )
        gen.generate_vds()

        self._validate_gap_fill(FEMS, CHIPS)

    def test_gap_fill_empty(self):
        FEMS = 6
        CHIPS = 8
        FRAMES = 100
        HEIGHT = 1536
        WIDTH = 2048
        print("Creating VDS...")
        gen = ExcaliburGapFillVDSGenerator(
            "./", files=["raw_0.h5"],
            source=dict(shape=(FRAMES, HEIGHT, WIDTH), dtype="float32"),
            chip_spacing=3, module_spacing=123,
            modules=3, output="gaps.h5", log_level=1
        )
        gen.generate_vds()
        # Generate a single file with 100 2048x1536 frames
        print("Creating raw files...")
        generate_raw_files("raw", FRAMES, 1, 1, WIDTH, HEIGHT)

        self._validate_gap_fill(FEMS, CHIPS)

    def _validate_gap_fill(self, fems, chips):
        print("Opening VDS...")
        with h5.File("gaps.h5", mode="r") as h5_file:
            vds_dataset = h5_file["data"]
            print("Verifying dataset...")
            # Check shape
            self.assertEqual(vds_dataset.shape, (100, 1791, 2069))

            self.assertEqual(vds_dataset[0][0][0], 0.0)  # FEM 1 pixel 1

            # Check corners where 4 chips meet have expected pattern
            row = 255
            column = 255
            spacings = [3, 123, 3, 123, 3]
            frame = vds_dataset[0]
            for stripe_idx in range(fems - 1):  # FEMS - 1 Vertical Gaps
                spacing = spacings[stripe_idx]
                pattern = create_pixel_pattern(
                    top_value=0.0, bottom_value=0.0,
                    horizontal_gap=spacing, vertical_gap=3, fill_value=-1.0
                )
                for chip_idx in range(chips - 1):  # CHIPS - 1 Horizontal Gaps
                    test = frame[row: row + spacing + 2, column: column + 5]
                    np.testing.assert_array_equal(pattern, test)
                    column += 256 + 3
                column = 255
                row += 256 + spacing

    def test_reshape(self):
        FRAMES = 60
        SHAPE = (5, 4, 3)
        # Generate a single file with 100 2048x1536 frames
        print("Creating raw files...")
        generate_raw_files("raw", FRAMES, 1, 1, 2048, 1536)
        print("Creating VDS...")
        gen = ReshapeVDSGenerator(
            shape=SHAPE, path="./", files=["raw_0.h5"],
            output="reshaped.h5", log_level=1
        )
        gen.generate_vds()

        self._validate_reshape(FRAMES)

    def test_reshape_alternate(self):
        FRAMES = 12
        SHAPE = (4, 3)
        # Generate a single file with 100 2048x1536 frames
        print("Creating raw files...")
        generate_raw_files("raw", FRAMES, 1, 1, 2048, 1536)
        print("Creating VDS...")
        gen = ReshapeVDSGenerator(
            shape=SHAPE, path="./", files=["raw_0.h5"],
            output="reshaped.h5", log_level=1,
            alternate=(False, True)
        )
        gen.generate_vds()

        test_dataset = np.array([0., 1., 2.,
                                 5., 4., 3.,
                                 6., 7., 8.,
                                 11., 10., 9.])

        with h5.File("reshaped.h5", mode="r") as h5_file:
            vds_dataset = h5_file["data"]

            print("Verifying dataset...")
            np.testing.assert_array_equal(
                test_dataset,
                vds_dataset[:, :, 0, 0].flatten()
            )

    def test_reshape_empty(self):
        FRAMES = 60
        SHAPE = (5, 4, 3)
        WIDTH = 2048
        HEIGHT = 1536
        print("Creating VDS...")
        gen = ReshapeVDSGenerator(
            shape=SHAPE, path="./", files=["raw_0.h5"],
            source=dict(shape=(FRAMES, HEIGHT, WIDTH), dtype="float32"),
            output="reshaped.h5", log_level=1
        )
        gen.generate_vds()
        # Generate a single file with 100 2048x1536 frames
        print("Creating raw files...")
        generate_raw_files("raw", FRAMES, 1, 1, 2048, 1536)

        self._validate_reshape(FRAMES)

    def test_reshape_speed(self):
        FRAMES = 1000000
        SHAPE = (500, 200, 10)
        WIDTH = 2048
        HEIGHT = 1536
        start = timer()
        print("Creating VDS...")
        gen = ReshapeVDSGenerator(
            shape=SHAPE, path="./", files=["raw_0.h5"],
            source=dict(shape=(FRAMES, HEIGHT, WIDTH), dtype="float32"),
            output="reshaped.h5", log_level=1
        )
        gen.generate_vds()
        end = timer()
        print("Reshape {} -> {} - Time: {}".format(FRAMES, SHAPE, end - start))

        start = timer()
        with h5.File("reshaped.h5", mode="r") as h5_file:
            dset = h5_file["data"]
        end = timer()
        print("Dataset open time: {}".format(end - start))

    def test_reshape_alternate_speed(self):
        FRAMES = 10000
        SHAPE = (50, 20, 10)
        WIDTH = 2048
        HEIGHT = 1536
        start = timer()
        print("Creating VDS...")
        gen = ReshapeVDSGenerator(
            shape=SHAPE, path="./", files=["raw_0.h5"],
            source=dict(shape=(FRAMES, HEIGHT, WIDTH), dtype="float32"),
            output="reshaped.h5", log_level=1, alternate=(False, True, True)
        )
        gen.generate_vds()
        end = timer()
        print("Reshape {} -> {} - Time: {}".format(FRAMES, SHAPE, end - start))

        start = timer()
        with h5.File("reshaped.h5", mode="r") as h5_file:
            dset = h5_file["data"]
        end = timer()
        print("Dataset open time: {}".format(end - start))

    def _validate_reshape(self, frames):
        print("Opening VDS...")
        with h5.File("reshaped.h5", mode="r") as h5_file:
            vds_dataset = h5_file["data"]

            print("Verifying dataset...")
            np.testing.assert_array_equal(
                np.array(range(frames)),
                vds_dataset[:, :, :, 0, 0].flatten()
            )
